#!/usr/bin/env python3.5
# -*- coding: utf-8 -*-
from .main import PrimaryNodesLocator


def test_locate_single_root_line():
    code = PrimaryNodesLocator('root').locate()

    num_line, line = next(code['root'])
    assert num_line == 1
    assert line == 'root'


def test_deny_bad_indent():
    code = '''
root:
 node'''
    code = PrimaryNodesLocator(code).locate()
    try:
        next(code['root'])
        next(code['root'])
        raise Exception('accepted bad indent')
    except SyntaxError:
        pass


def test_deny_multiroot():
    code = '''
node
node'''
    try:
        PrimaryNodesLocator(code).locate()
        raise Exception('accepted multiroot')
    except SyntaxError:
        pass


def test_multiline():
    code = '''
root: |
    hello
    world
'''
    code = PrimaryNodesLocator(code).locate()

    for num_line, line in code['root']:
        if num_line == 2:
            assert line == 'root: hello\nworld'
        elif num_line < 2 or num_line > 2:
            error(num_line, line)


def error(num_line, line):
    raise AssertionError('{}:{}'.format(num_line, line))


def test_locate_multiline_root_and_rules():
    code = '''
root:
    node1
    node2
<rule1@x>:
    node3
<rule2@x>:
    node4:
        node5
'''
    code = PrimaryNodesLocator(code).locate()
    check_root(code)
    check_first_rule(code)
    check_second_rule(code)


def check_root(code):
    for num_line, line in code['root']:
        if num_line == 2:
            assert line == 'root:'
        elif num_line == 3:
            assert line == '    node1'
        elif num_line == 4:
            assert line == '    node2'
        elif num_line < 2 or num_line > 4:
            error(num_line, line)


def check_first_rule(code):
    for num_line, line in code['rules']['rule1']:
        if num_line == 5:
            assert line == '<rule1@x>:'
        elif num_line == 6:
            assert line == '    node3'
        elif num_line < 5 or num_line > 6:
            error(num_line, line)


def check_second_rule(code):
    for num_line, line in code['rules']['rule2']:
        if num_line == 7:
            assert line == '<rule2@x>:'
        elif num_line == 8:
            assert line == '    node4:'
        elif num_line == 9:
            assert line == '        node5'
        elif num_line < 7 or num_line > 9:
            error(num_line, line)


def test_ignore_comment():
    code = '''
node
    #comment_1
# comment_2
 #comment_3'''
    code = PrimaryNodesLocator(code).locate()

    i = 0
    for a, b in code['root']:
        i += 1
    assert i == 1


def test_raw_text():
    code = PrimaryNodesLocator('## raw text').locate()
    num_line, line = code['raw'][0]
    assert num_line == 1
    assert line == 'raw text'



#!/usr/bin/env python3.5
# -*- coding: utf-8 -*-
from ..node import TreeBuilder, PrimaryNodesLocator
from kivish.utils import error


class Scanner:
    '''Scans a kvs code to build a tree of nodes.'''

    def __init__(self, code):
        self._code = code

    def scan(self):
        '''
        Builds a tree of nodes. Returns a root node and a list
        of raw text lines.
        '''
        r = PrimaryNodesLocator(self._code).locate()
        root_tree, rules, raw = r['root'], r['rules'], r['raw']

        if not root_tree:
            error.report('no root node found')

        root = TreeBuilder(root_tree, rules).go()
        return {'root': root, 'raw': raw}



#!/usr/bin/env python3.5
# -*- coding: utf-8 -*-
import re

_name = '[A-Za-z0-9-]+'
_star_or_slash = '(?:\*?|\/?)'

pattern_label = re.compile('^{} *({})$'.format(_star_or_slash, _name))
pattern_class = re.compile('\.({})'.format(_name))
pattern_id = re.compile('#({})'.format(_name))



#!/usr/bin/env python3.5
# -*- coding: utf-8 -*-
from .config import config
INPUT_INDENT = config['INPUT_INDENT']
OUTPUT_INDENT = config['OUTPUT_INDENT']

from kivish.utils import error


def is_inline_node(node):
    return node.label[0] == '/'


def get_spaces(indent: int) -> str:
    if indent == 0:
        return ''
    else:
        num_tabs = int(indent / INPUT_INDENT)
        return ' ' * (OUTPUT_INDENT * num_tabs)


def is_attribute_node(node):
    if node.label[0] == '*':
        if node.children:
            msg = 'attribute node cannot have children'
            error.report(msg, node.num_line)
        return True


def get_attribute(child):
    if child.value:
        child.value = child.value.replace('\n', ' ')
        return ' ' + child.label[1:] + '="' + child.value + '"'
    else:
        return ' ' + child.label[1:]


def bump_attribute_nodes(children):
    '''Move attribute nodes to the top of children list.'''
    que = []

    for child in children:
        if child.label[0] == '*':
            yield child
        else:
            que.append(child)

    for child in que:
        yield child


def embed_attribute(child, string):
    string = string[:-2]
    if child.value:
        child.value = child.value.replace('\n', ' ')
        string += ' ' + child.label[1:] + '="' + child.value + '">\n'
    else:
        string += ' ' + child.label[1:] + '>\n'
    return string


def add_value(string, node, spaces):
    if node.value:

        parts = node.value.split('\n')
        for part in parts:
            string += spaces + ' ' * OUTPUT_INDENT + part + '\n'

    return string


def is_raw_text_node(node):
    return node.label == '--RAW'


def create_raw_text_node(node):
    return get_spaces(node.indent) + node.value + '\n'



#!/usr/bin/env python3.5
# -*- coding: utf-8 -*-
from kivish.utils import error, regex
from . import utils


class Generator:
    '''Creates an html file from a provided kvs root node.'''

    def __init__(self, root_node, raw_text=None):
        self._root_node = root_node
        self._raw_text = raw_text

    def generate(self):
        '''Creates an html string and returns it.'''
        html = self._node_to_string(self._root_node)

        while self._raw_text:
            html += self._pop_raw_text()

        return html

    def _node_to_string(self, node):
        string = ''

        while self._should_put_raw_text(node):
            string += self._pop_raw_text()

        if utils.is_inline_node(node):
            string += self._create_inline_node(node)
        elif utils.is_raw_text_node(node):
            string += utils.create_raw_text_node(node)
        else:
            string += self._create_normal_node(node)

        return string

    def _should_put_raw_text(self, node):
        try:
            return node.num_line > self._raw_text[0][0]
        except:
            return False

    def _pop_raw_text(self):
        string = self._raw_text[0][1]
        del self._raw_text[0]
        return string + '\n'

    def _create_inline_node(self, node):
        spaces = utils.get_spaces(node.indent)
        label = regex.pattern_label.findall(node.label)[0]

        string = spaces + '<' + label

        for child in node.children:
            if utils.is_attribute_node(child):
                string += utils.get_attribute(child)
            else:
                msg = 'inline node can have only attribute children'
                error.report(msg, node.num_line)

        string += '/>\n'
        return string

    def _create_normal_node(self, node):
        spaces = utils.get_spaces(node.indent)
        label = regex.pattern_label.findall(node.label)[0]

        string = spaces + '<' + label + '>\n'

        for child in utils.bump_attribute_nodes(node.children):
            if utils.is_attribute_node(child):
                string = utils.embed_attribute(child, string)
            else:
                string += self._node_to_string(child)

        string = utils.add_value(string, node, spaces)

        string += spaces + '</' + label + '>\n'
        return string
